#! /usr/bin/env python
"""The module creates some basic constants to describe the Pyslet package."""

title_name = u"Pyslet"
name = "pyslet"
copyright = u"\xA92008-2017, Steve Lay"

major_version = "0.7"
build_date = "20170805"
version = "%s.%s" % (major_version, build_date)

title = (
    "Pyslet: "
    "Python package for Standards in Learning, Education and Training")

home = "http://www.pyslet.org/"
